use anyhow::{Ok, Result};
use bytesize::ByteSize;
use clap::Parser;
use indicatif::ProgressBar;
use std::process::Command;
use std::{
    collections::HashMap,
    fs,
    path::{Path, PathBuf},
    vec,
};
use temp_dir::TempDir;

use strum::IntoEnumIterator;
use strum_macros::EnumIter;

#[derive(Parser, Debug)]
#[command(version, about, long_about = "This is a small tool I've made to make quickly comparing compression formats easier!\n\nUse `--lzma` to just test lzma or use multiple specifiers to just test them (`--lzma --bzip3`)\n\nThe commandline flags used to test are listed bellow.")]
struct Args {
    /// Path of file for compression tests
    #[arg()]
    path: PathBuf,

    /// Test Bzip2 compression (--best -c) - Tested by default
    #[arg(long)]
    bzip2: bool,

    /// Test Bzip3 compression (-e -c) - Tested by default
    #[arg(long)]
    bzip3: bool,

    /// Test LZMA compression (-z -c -9)
    #[arg(long)]
    lzma: bool,

    /// Test LZMA compression (-z -c -9e)
    #[arg(long)]
    lzma_e: bool,

    /// Test ZSTD compression (-z -9 -T0 -c)
    #[arg(long)]
    zstd9: bool,

    /// Test ZSTD compression (-z -19 -T0 -c) - Tested by default
    #[arg(long)]
    zstd19: bool,

    /// Test ZSTD compression (-z --ultra -22 -T0 -c)
    #[arg(long)]
    zstd22: bool,

        /// Test all formats
        #[arg(long, short)]
        all: bool,
}

#[derive(Parser, Debug, PartialEq, Eq, Hash, EnumIter)]
enum Formats {
    LZMA,
    LzmaE,
    Bzip2,
    Bzip3,
    ZSTD9,
    ZSTD19,
    ZSTD22,
}

fn main() -> Result<(), anyhow::Error> {
    let args = Args::parse();

    let mut tests = vec![Formats::Bzip2, Formats::Bzip3, Formats::ZSTD19];

    if args.all {
        tests = vec![];
        for format in Formats::iter() {
            tests.push(format);
        }
    } else if args.bzip2 || args.bzip3 || args.lzma || args.lzma_e || args.zstd9 || args.zstd19 || args.zstd22 {
        tests = vec![];

        if args.bzip2 {
            tests.push(Formats::Bzip2);
        }

        if args.bzip3 {
            tests.push(Formats::Bzip3);
        }

        if args.lzma {
            tests.push(Formats::LZMA);
        }

        if args.lzma_e {
            tests.push(Formats::LzmaE);
        }

        if args.zstd9 {
            tests.push(Formats::ZSTD9);
        }

        if args.zstd19 {
            tests.push(Formats::ZSTD19);
        }

        if args.zstd22 {
            tests.push(Formats::ZSTD22);
        }
    }

    let temp_dir = TempDir::new()?;

    let mut results: HashMap<Formats, u64> = HashMap::new();

    let bar = ProgressBar::new(tests.len() as u64);
    bar.tick();

    for test in tests {
        let file_size = match test {
            Formats::LZMA => lzma_test(args.path.clone(), temp_dir.path())?,
            Formats::Bzip2 => bzip2_test(args.path.clone(), temp_dir.path())?,
            Formats::Bzip3 => bzip3_test(args.path.clone(), temp_dir.path())?,
            Formats::ZSTD9 => zstd9_test(args.path.clone(), temp_dir.path())?,
            Formats::ZSTD19 => zstd19_test(args.path.clone(), temp_dir.path())?,
            Formats::ZSTD22 => zstd22_test(args.path.clone(), temp_dir.path())?,
            Formats::LzmaE => lzma_e_test(args.path.clone(), temp_dir.path())?,
        };

        results.insert(test, file_size);
        bar.inc(1);
    }
    bar.finish();

    for (format, size) in results {
        println!("{:?}: {}", format, ByteSize::b(size).display().si())
    }

    temp_dir.cleanup()?;

    Ok(())
}

fn bzip2_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "bzip2 --best -c {} > {}/{}.bzip2",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "bzip2 --best -c {} > {}/{}.bzip2",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}.bzip2",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}

fn bzip3_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "bzip3 -e -c {} > {}/{}.bzip3",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "bzip3 -e -c {} > {}/{}.bzip3",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}.bzip3",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}

fn zstd9_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "zstd -z -9 -T0 -c {} > {}/{}-9.zstd",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "zstd -z -9 -T0 -c {} > {}/{}-9.zstd",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}-9.zstd",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}

fn zstd19_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "zstd -z -19 -T0 -c {} > {}/{}-19.zstd",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "zstd -z -19 -T0 -c {} > {}/{}-19.zstd",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}-19.zstd",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}

fn zstd22_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "zstd -z --ultra -22 -T0 -c {} > {}/{}-22.zstd",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "zstd -z --ultra -22 -T0 -c {} > {}/{}-22.zstd",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}-22.zstd",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}

fn lzma_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "xz -z -c -9 {} > {}/{}.xz",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "xz -z -c -9 {} > {}/{}.xz",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}.xz",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}

fn lzma_e_test(path: PathBuf, temp_dir: &Path) -> Result<u64, anyhow::Error> {
    let _output = if cfg!(target_os = "windows") {
        Command::new("cmd")
            .arg("/C")
            .arg(format!(
                "xz -z -c -9e {} > {}/{}.xz",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    } else {
        Command::new("sh")
            .arg("-c")
            .arg(format!(
                "xz -z -c -9e {} > {}/{}.xz",
                path.display(),
                temp_dir.display(),
                path.file_name().unwrap().to_str().unwrap()
            ))
            .output()
            .expect("failed to execute process")
    };

    let metadata = fs::metadata(format!(
        "{}/{}.xz",
        temp_dir.display(),
        path.file_name().unwrap().to_str().unwrap()
    ))?;

    Ok(metadata.len())
}
